"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.where = exports.betweenDate = exports.betweenNumber = exports.between = void 0;
const lodash_1 = __importDefault(require("lodash"));
const Util_1 = require("../base/Util");
const dayjs_1 = __importDefault(require("dayjs"));
const BOOLEAN_TEXT_IGNORE = new Set(['', 'null']);
const BOOLEAN_TEXT_FALSE = new Set(['0', 'false', '-1']);
const NullCondition = (item, pos) => {
    if (!Util_1.NONE_PARAM.has(item.fn))
        return false;
    let bool = true;
    if (lodash_1.default.isString(item.value)) {
        let v = lodash_1.default.toLower(lodash_1.default.trim(item.value));
        if (BOOLEAN_TEXT_IGNORE.has(v))
            return;
        if (BOOLEAN_TEXT_FALSE.has(v))
            bool = false;
    }
    else {
        bool = !!item.value;
    }
    switch (item.fn) {
        case 'IsNull':
            pos.SQL.push(`${item.column} IS ${bool ? '' : 'NOT'} NULL`);
            return true;
        case 'NotNull':
            pos.SQL.push(`${item.column} IS ${bool ? 'NOT' : ''} NULL`);
            return true;
        case 'IsDistinct':
            pos.SQL.push(`${item.column} IS ${bool ? '' : 'NOT'} DISTINCT`);
            return true;
        case 'NotDistinct':
            pos.SQL.push(`${item.column} IS ${bool ? 'NOT' : ''} DISTINCT`);
            return true;
    }
};
const compareSuffix = (suffix) => {
    switch (suffix) {
        case '<':
        case '>':
        case '<=':
        case '>=':
        case '=':
            return suffix;
        case 'Min':
            return '>';
        case 'Max':
            return '<';
        case 'MinD':
        case 'MinH':
        case 'MinM':
        case 'MinThan':
            return '>=';
        case 'MaxD':
        case 'MaxH':
        case 'MaxM':
        case 'MaxThan':
            return '<=';
        case 'Not':
        case '!=':
        case '<>':
            return '!=';
    }
    return null;
};
const getStart = (str) => {
    if (str.startsWith('('))
        return ['>', str.substring(1)];
    if (str.startsWith('['))
        return ['>=', str.substring(1)];
    return ['>=', str];
};
const getEnd = (str) => {
    if (str.endsWith(')'))
        return ['<', str.substring(0, str.length - 1)];
    if (str.endsWith(']'))
        return ['<=', str.substring(0, str.length - 1)];
    return ['<=', str];
};
const between = (txt) => {
    let str = lodash_1.default.trim((txt || ''));
    if (str.length == 0)
        return null;
    let ptns = str.split(',').map(lodash_1.default.trim);
    if (ptns.length > 2)
        return null;
    if (ptns.length == 1)
        return [getStart(ptns[0])];
    let result = [];
    if (ptns[0].length)
        result.push(getStart(ptns[0]));
    if (ptns[1].length)
        result.push(getEnd(ptns[1]));
    return result;
};
exports.between = between;
const betweenNumber = (txt) => {
    let range = (0, exports.between)(txt);
    if (range == null)
        return null;
    let result = [];
    for (let item of range) {
        try {
            let num = parseFloat(item[1]);
            result.push([item[0], num]);
        }
        catch (_a) {
            return null;
        }
    }
    return result;
};
exports.betweenNumber = betweenNumber;
const betweenDate = (txt) => {
    let range = (0, exports.between)(txt);
    if (range == null)
        return null;
    let result = [];
    for (let item of range) {
        let day = (0, dayjs_1.default)(item[1]);
        if (!day.isValid()) {
            return null;
        }
        result.push([item[0], day.toDate()]);
    }
    return result;
};
exports.betweenDate = betweenDate;
const whereText = (item, pos, err) => {
    if (NullCondition(item, pos))
        return;
    const compare = compareSuffix(item.fn);
    if (compare != null) {
        pos.SQL.push(`${item.column} ${compare} $${pos.NUM}`);
        pos.PARAM.push(item.value);
        pos.NUM++;
        return;
    }
    switch (item.fn) {
        case 'Like':
            pos.SQL.push(`${item.column} LIKE $${pos.NUM}`);
            pos.PARAM.push('%' + item.value + '%');
            pos.NUM++;
            return;
        case 'Likel':
            pos.SQL.push(`${item.column} LIKE $${pos.NUM}`);
            pos.PARAM.push(item.value + '%');
            pos.NUM++;
            return;
        case 'Liker':
            pos.SQL.push(`${item.column} LIKE $${pos.NUM}`);
            pos.PARAM.push('%' + item.value);
            pos.NUM++;
            return;
        default:
            err.push(`${item.column}/ type : String not support method ${item.fn}`);
            return;
    }
};
const whereNumber = (item, pos, err) => {
    if (NullCondition(item, pos))
        return;
    const compare = compareSuffix(item.fn);
    if (compare != null) {
        try {
            let val = lodash_1.default.isNumber(item.value) ? item.value : parseFloat(item.value);
            pos.SQL.push(`${item.column} ${compare} $${pos.NUM}`);
            pos.PARAM.push(val);
            pos.NUM++;
        }
        catch (_a) {
            err.push(`${item.column}/ type : Number value is not a Number ${item.value}`);
        }
        return;
    }
    if (item.fn == 'Bt') {
        let range = (0, exports.betweenNumber)(item.value + '');
        if (range == null) {
            err.push(`${item.column}/(Number) :  Between Value invalidated ${item.value}`);
            return;
        }
        range.map(oper => {
            pos.SQL.push(`${item.column} ${oper[0]} $${pos.NUM}`);
            pos.PARAM.push(oper[1]);
            pos.NUM++;
        });
        return;
    }
    err.push(`${item.column}/(Number) : not support method ${item.fn}`);
};
const whereDate = (item, pos, err) => {
    if (NullCondition(item, pos))
        return;
    let val = null;
    if (item.fn != 'Bt') {
        if (item.value == '' || item.value == null) {
            err.push(`${item.column}/(date) : Can not be null`);
            return;
        }
        val = (0, dayjs_1.default)(item.value);
        if (!val.isValid()) {
            err.push(`${item.column}/(date) : Must Be a date-string or number-stamp ${item.value}`);
            return;
        }
        switch (item.fn) {
            case 'MinH':
                val = val.startOf('hour');
                break;
            case 'MinD':
                val = val.startOf('date');
                break;
            case 'MinM':
                val = val.startOf('month');
                break;
            case 'MaxH':
                val = val.endOf('month');
                break;
            case 'MaxD':
                val = val.endOf('month');
                break;
            case 'MaxM':
                val = val.endOf('month');
                break;
        }
    }
    const compare = compareSuffix(item.fn);
    if (compare != null) {
        pos.SQL.push(`${item.column} ${compare} $${pos.NUM}`);
        pos.PARAM.push(val.toDate());
        pos.NUM++;
        return;
    }
    let start = null, end = null;
    switch (item.fn) {
        case 'BtD':
            start = val.clone().startOf('date').toDate();
            end = val.clone().endOf('date').toDate();
            break;
        case 'BtM':
            start = val.clone().startOf('month').toDate();
            end = val.clone().endOf('month').toDate();
            break;
        case 'BtY':
            start = val.clone().startOf('year').toDate();
            end = val.clone().endOf('year').toDate();
            break;
        case 'Bt':
            const range = (0, exports.betweenDate)(item.value + '');
            if (range == null) {
                err.push(`${item.column}/(Date) :  Between Value invalidated ${item.value}`);
                return;
            }
            range.map(oper => {
                pos.SQL.push(`${item.column} ${oper[0]} $${pos.NUM}`);
                pos.PARAM.push(oper[1]);
                pos.NUM++;
            });
            return;
    }
    if (start == null || end == null) {
        err.push(`${item.column}/(Date) : not support method ${item.fn}`);
        return;
    }
    pos.SQL.push(`${item.column} >= $${pos.NUM}`);
    pos.PARAM.push(start);
    pos.NUM++;
    pos.SQL.push(`${item.column} <= $${pos.NUM}`);
    pos.PARAM.push(end);
    pos.NUM++;
};
const whereBoolean = (item, pos, err) => {
    if (NullCondition(item, pos))
        return;
    let bool = true;
    if (lodash_1.default.isString(item.value)) {
        let v = lodash_1.default.toLower(lodash_1.default.trim(item.value));
        if (BOOLEAN_TEXT_IGNORE.has(v))
            return;
        if (BOOLEAN_TEXT_FALSE.has(v))
            bool = false;
    }
    else {
        bool = !!item.value;
    }
    switch (item.fn) {
        case 'IsNull':
            pos.SQL.push(`${item.column} IS ${bool ? '' : 'NOT'} NULL`);
            return;
        case 'NotNull':
            pos.SQL.push(`${item.column} IS ${bool ? 'NOT' : ''} NULL`);
            return;
        case '!=':
        case '<>':
        case 'Not':
        case 'Min':
        case 'Max':
        case '>':
        case '<':
            bool = !bool;
            break;
        default:
            break;
    }
    pos.SQL.push(`${item.column} IS ${bool ? '' : 'NOT'} TRUE`);
};
const ItemToWhere = (whereItem, pos, err) => {
    let item = Object.assign(Object.assign({}, whereItem), { fn: whereItem.fn ? whereItem.fn : '=', type: whereItem.type ? whereItem.type : 'string' });
    if (!(0, Util_1.isSupport)(item.type, item.fn)) {
        err.push(`${item.column}/(${item.type}) not support method ${item.fn}`);
        return;
    }
    switch (item.type) {
        case 'number':
            whereNumber(item, pos, err);
            return;
        case 'string':
            whereText(item, pos, err);
            return;
        case 'boolean':
            whereBoolean(item, pos, err);
            return;
        case 'date':
            whereDate(item, pos, err);
            return;
        default:
            whereText(item, pos, err);
            return;
    }
};
const ConditionToWhere = (condition, pos, err) => {
    for (let item of condition.items) {
        if (lodash_1.default.has(item, 'link')) {
            let group = item;
            let _pos = {
                SQL: [],
                PARAM: [],
                NUM: pos.NUM
            };
            ConditionToWhere(group, _pos, err);
            if (_pos.SQL.length) {
                if (_pos.NUM > pos.NUM)
                    pos.NUM = _pos.NUM;
                for (let param of _pos.PARAM)
                    pos.PARAM.push(param);
                // let link = group.link == 'NOT' ? 'AND NOT' : group.link;
                pos.SQL.push('(' + _pos.SQL.join(' ' + group.link + ' ') + ')');
            }
        }
        else {
            ItemToWhere(item, pos, err);
        }
    }
};
const where = (condition, startIdx = 1) => {
    const pos = { SQL: [], PARAM: [], NUM: startIdx };
    let root = lodash_1.default.isArray(condition) ? { link: 'AND', items: condition } : condition;
    let err = [];
    ConditionToWhere(root, pos, err);
    (0, Util_1.throwErr)(err, 'Some SQL Error Occur');
    if (pos.SQL.length == 0) {
        return ['', []];
    }
    ;
    return [pos.SQL.join(" " + root.link + " "), pos.PARAM];
};
exports.where = where;
